/*
 * Copyright (c) 2005-2006 Erez Zadok,
 * Copyright (c) 2005-2006 Nikolai Joukov,
 * Copyright (c) 2005      Arun M. Krishnakumar,
 * Copyright (c) 2006      Chaitanya Patti,
 * Copyright (c) 2005      Sunil Satnur,
 * Copyright (c) 2006      Avishay Traeger,
 * Copyright (c) 2005-2006 Stony Brook University
 *
 * For specific licensing information, see the COPYING file distributed with
 * this package, or get one from ftp://ftp.filesystems.org/pub/fist/COPYING.
 *
 * This Copyright notice must be kept intact and distributed with all
 * fistgen sources INCLUDING sources generated by fistgen.
 */
/*
 *  $Id: migrate.c,v 1.8 2006/12/31 01:43:58 crpatti Exp $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include "fist.h"
#include "raif.h"

int raif_mig_unlink(char *pathname)
{
        int error = 0;
        struct dentry *dentry;
        struct nameidata nd;
        struct inode *inode = NULL;

        error = path_lookup(pathname, LOOKUP_PARENT, &nd);
        if (error)
                goto exit;
        error = -EISDIR;
        if (nd.last_type != LAST_NORM)
                goto exit1;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,16)
        down(&nd.dentry->d_inode->i_sem);
#else
        mutex_lock(&nd.dentry->d_inode->i_mutex);
#endif
	dentry = lookup_one_len(nd.last.name, nd.dentry, (int)nd.last.len );
        error = PTR_ERR(dentry);
        if (!IS_ERR(dentry)) {
                /* Why not before? Because we want correct error value */
                if (nd.last.name[nd.last.len])
                        goto slashes;
                inode = dentry->d_inode;
                if (inode)
                        atomic_inc(&inode->i_count);
                error = vfs_unlink(nd.dentry->d_inode, dentry);
        exit2:
                dput(dentry);
        }
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,16)
        up(&nd.dentry->d_inode->i_sem);
#else
        mutex_unlock(&nd.dentry->d_inode->i_mutex);
#endif
        if (inode)
                iput(inode);    /* truncate the inode here */
exit1:
        path_release(&nd);
exit:
        return error;

slashes:
        error = !dentry->d_inode ? -ENOENT :
                S_ISDIR(dentry->d_inode->i_mode) ? -EISDIR : -ENOTDIR;
        goto exit2;

}

int raif_mig_rename( char *oldname, char *newname)
{
        int error = 0;
        struct dentry * old_dir, * new_dir;
        struct dentry * old_dentry, *new_dentry;
        struct dentry * trap;
        struct nameidata oldnd, newnd;

        error = path_lookup(oldname, LOOKUP_PARENT, &oldnd);
        if (error)
                goto exit;

        error = path_lookup(newname, LOOKUP_PARENT, &newnd);
        if (error)
                goto exit1;

        error = -EXDEV;
        if (oldnd.mnt != newnd.mnt)
                goto exit2;

        old_dir = oldnd.dentry;
        error = -EBUSY;
        if (oldnd.last_type != LAST_NORM)
                goto exit2;

        new_dir = newnd.dentry;
        if (newnd.last_type != LAST_NORM)
                goto exit2;

        trap = lock_rename(new_dir, old_dir);

	old_dentry = lookup_one_len(oldnd.last.name, old_dir, (int)oldnd.last.len );
        error = PTR_ERR(old_dentry);
        if (IS_ERR(old_dentry))
                goto exit3;
        /* source must exist */
        error = -ENOENT;
        if (!old_dentry->d_inode)
                goto exit4;
        /* unless the source is a directory trailing slashes give -ENOTDIR */
        if (!S_ISDIR(old_dentry->d_inode->i_mode)) {
                error = -ENOTDIR;
                if (oldnd.last.name[oldnd.last.len])
                        goto exit4;
                if (newnd.last.name[newnd.last.len])
                        goto exit4;
        }
        /* source should not be ancestor of target */
        error = -EINVAL;
        if (old_dentry == trap)
                goto exit4;
	new_dentry = lookup_one_len(newnd.last.name, new_dir, (int)newnd.last.len );
        error = PTR_ERR(new_dentry);
        if (IS_ERR(new_dentry))
                goto exit4;
        /* target should not be an ancestor of source */
        error = -ENOTEMPTY;
        if (new_dentry == trap)
                goto exit5;

        error = vfs_rename(old_dir->d_inode, old_dentry,
                                   new_dir->d_inode, new_dentry);
exit5:
        dput(new_dentry);
exit4:
        dput(old_dentry);
exit3:
        unlock_rename(new_dir, old_dir);
exit2:
        path_release(&newnd);
exit1:
        path_release(&oldnd);
exit:
        return error;

}

int raif_migrate(file_t *dirfile)
{
	dentry_t *dir = dirfile->f_dentry;
	file_t *src, *dst;
	mm_segment_t oldfs;
	struct rule *localrule, *ruletemp;
	struct raif_dentry_info *rdi = dtopd(dir);
	char *tmpname = NULL, *buff = NULL;
	int err = 0;
	long nbytes;

	write_lock(&rdi->d.rules.spin);
	oldfs= get_fs();
	localrule = rdi->d.rules.local_rule;
	while(1) {
		if (!localrule) { 
			goto out_free;
		}

		src = filp_open(localrule->regexp, O_RDONLY, 0);
		if (IS_ERR(src)) {
			err = PTR_ERR(src);
			localrule = localrule->next;
			continue;
		}

                tmpname = KMALLOC(strlen(localrule->regexp) + strlen(RAIF_TMPREFIX) + 1, GFP_KERNEL);
                if (!tmpname) {
                        err = -ENOMEM;
                        goto out_free;
                }
                strcpy(tmpname, RAIF_TMPREFIX);
		strcat(tmpname, localrule->regexp);

		dst = filp_open(tmpname, O_RDWR | O_CREAT, 0);
		if (IS_ERR(dst)) {
			err = PTR_ERR(dst);
			goto out_src_close;
		}

		buff = KMALLOC(PAGE_CACHE_SIZE, GFP_KERNEL);
		if (!buff) {
			err = -ENOMEM;
			goto out_close;
		}
		src->f_pos = 0;
		dst->f_pos = 0;

		set_fs(KERNEL_DS);

		do {
			nbytes = src->f_op->read(src, buff, PAGE_CACHE_SIZE, &src->f_pos);
			if (nbytes < 0) {
				err = nbytes;
				goto out_close;
			}
			dst->f_op->write(dst, buff, nbytes, &dst->f_pos);
		} while(nbytes);
		dst->f_dentry->d_inode->i_mode = src->f_dentry->d_inode->i_mode;
		dst->f_dentry->d_inode->i_uid = src->f_dentry->d_inode->i_uid;
		dst->f_dentry->d_inode->i_gid = src->f_dentry->d_inode->i_gid;
		set_fs(oldfs);

		filp_close(src, NULL);
		filp_close(dst, NULL);

		raif_mig_unlink(localrule->regexp);
		raif_mig_rename(tmpname, localrule->regexp);

		/* Now delete the local rule */
		ruletemp = rdi->d.rules.local_rule;
		if (ruletemp == localrule) {
			rdi->d.rules.local_rule = ruletemp->next;
		} else {
			while (ruletemp->next != localrule)
				ruletemp = ruletemp->next;
			ruletemp->next = localrule->next;
		}
		KFREE(localrule);
		localrule = ruletemp->next;
	}

out_close:
	filp_close(dst, NULL);
out_src_close:
	filp_close(src, NULL);
out_free:
	set_fs(oldfs);
	write_unlock(&rdi->d.rules.spin);
	if (tmpname)
		KFREE(tmpname);
	if (buff)
		KFREE(buff);
	return err;
}
